#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import commons

import appier
import appier_extras

from . import base
from . import bundle
from . import currency as _currency

class Product(base.BudyBase):

    GENDER_S = {
        "Male" : "Male",
        "Female" : "Female",
        "Child" : "Child",
        "Both" : "Both"
    }
    """ The dictionary that maps the multiple gender
    enumeration values with their string representation """

    short_description = appier.field(
        index = "hashed",
        default = True,
        observations = """A short description on the product
        that should be used as the title for the product"""
    )

    product_id = appier.field(
        index = "hashed",
        description = "Product ID",
        observations = """The primary identifier of the
        product, should be globally unique"""
    )

    supplier_code = appier.field(
        index = "hashed",
        observations = """The identifier used by the
        (back) supplier of the product, should be considered
        a more unique way of identifying a product"""
    )

    sku = appier.field(
        index = "hashed",
        description = "SKU",
        observations = """The main identifier to be used for
        the keeping of the internal reference SKU (Stock Keeping
        Unit), should be considered the public way of representing
        the product"""
    )

    upc = appier.field(
        index = "hashed",
        description = "UPC",
        observations = """The standard Universal Product Code
        (UPC) value for this product"""
    )

    ean = appier.field(
        index = "hashed",
        description = "EAN",
        observations = """The standard European Article Number
        (EAN) value for this product"""
    )

    gender = appier.field(
        index = "hashed",
        meta = "enum",
        enum = GENDER_S
    )

    weight = appier.field(
        type = commons.Decimal,
        index = True,
        observations = """The weight of the current product in
        a unit defined by convention (defined before-hand)"""
    )

    quantity_hand = appier.field(
        type = commons.Decimal,
        index = True,
        observations = """The total quantity that is currently
        available (on hand) for the current product, if not
        set the product is considered to have unlimited quantity
        (inventory is not controller)"""
    )

    quantity_reserved = appier.field(
        type = commons.Decimal,
        index = True,
        observations = """The total quantity that is currently
        reserved (not available) for the current product, if not
        set the product is considered to not have reservation
        control enabled"""
    )

    price = appier.field(
        type = commons.Decimal,
        index = True,
        initial = commons.Decimal(0.0),
        observations = """Main retail price to be used for
        a possible sale transaction of the product (includes taxes)"""
    )

    price_compare = appier.field(
        type = commons.Decimal,
        index = True,
        initial = commons.Decimal(0.0),
        observations = """The price that is going to be used
        as the base for discount calculation purposes"""
    )

    taxes = appier.field(
        type = commons.Decimal,
        index = True,
        initial = commons.Decimal(0.0),
        observations = """The amount of taxes that are associated
        with the product this value can include the VAT value but
        it's not limited to it and more tax values can be included"""
    )

    currency = appier.field(
        index = "hashed"
    )

    order = appier.field(
        type = int,
        index = "hashed"
    )

    tag = appier.field()

    tag_description = appier.field()

    price_provider = appier.field(
        index = True
    )

    discountable = appier.field(
        type = bool,
        initial = True,
        observations = """Flag that indicates if the product is
        eligible for any kind of global discount"""
    )

    orderable = appier.field(
        type = bool,
        observations = """Flag that controls if the product is only
        meant to be used for pre-ordering and not direct sales"""
    )

    price_url = appier.field(
        index = "hashed",
        meta = "url",
        description = "Price URL"
    )

    farfetch_url = appier.field(
        index = "hashed",
        meta = "url",
        description = "Farfetch URL"
    )

    farfetch_male_url = appier.field(
        index = "hashed",
        meta = "url",
        description = "Farfetch Male URL"
    )

    farfetch_female_url = appier.field(
        index = "hashed",
        meta = "url",
        description = "Farfetch Female URL"
    )

    image_url = appier.field(
        index = "hashed",
        meta = "image_url",
        description = "Image URL"
    )

    thumbnail_url = appier.field(
        index = "hashed",
        meta = "image_url",
        description = "Thumbnail URL",
        observations = """The URL to the thumbnail image (low resolution)
        to be used in listing contexts"""
    )

    characteristics = appier.field(
        type = list,
        observations = """The sequence of characteristics that
        define the product under unstructured language"""
    )

    features = appier.field(
        type = dict,
        observations = """Dictionary that maps a set of optional
        features with their corresponding configuration, for example
        the `initials` feature can be associated with the rules
        that control those same initials"""
    )

    labels = appier.field(
        type = list,
        observations = """Set of simple tag labels that define the
        product behavior, this value is a calculated one based on
        the labels provided by the complete set of collections"""
    )

    brand_s = appier.field(
        index = "hashed"
    )

    season_s = appier.field(
        index = "hashed"
    )

    color_s = appier.field(
        index = "hashed"
    )

    category_s = appier.field(
        index = "hashed"
    )

    collection_s = appier.field(
        index = "hashed"
    )

    colors = appier.field(
        type = appier.references(
            "Color",
            name = "id"
        )
    )

    categories = appier.field(
        type = appier.references(
            "Category",
            name = "id"
        )
    )

    collections = appier.field(
        type = appier.references(
            "Collection",
            name = "id"
        )
    )

    variants = appier.field(
        type = appier.references(
            "Product",
            name = "id"
        )
    )

    images = appier.field(
        type = appier.references(
            "Media",
            name = "id"
        )
    )

    brand = appier.field(
        type = appier.reference(
            "Brand",
            name = "id"
        )
    )

    season = appier.field(
        type = appier.reference(
            "Season",
            name = "id"
        )
    )

    measurements = appier.field(
        type = appier.references(
            "Measurement",
            name = "id"
        )
    )

    compositions = appier.field(
        type = appier.references(
            "Composition",
            name = "id"
        )
    )

    live_model = appier.field(
        type = appier.references(
            "LiveModel",
            name = "id"
        )
    )

    @classmethod
    def validate(cls):
        return super(Product, cls).validate() + [
            appier.not_null("short_description"),
            appier.not_empty("short_description"),

            appier.not_null("gender"),
            appier.not_empty("gender"),

            appier.not_null("price"),
            appier.gte("price", 0.0),

            appier.not_null("taxes"),
            appier.gte("taxes", 0.0),

            appier.not_empty("brand_s"),

            appier.not_empty("season_s"),

            appier.not_empty("color_s"),

            appier.not_empty("category_s"),

            appier.not_empty("collection_s")
        ]

    @classmethod
    def list_names(cls):
        return ["id", "product_id", "short_description", "enabled", "gender", "tag", "image_url"]

    @classmethod
    def order_name(cls):
        return ["id", -1]

    @classmethod
    def index_names(cls):
        return super(Product, cls).index_names() + ["product_id"]

    @classmethod
    def token_names(cls):
        return super(Product, cls).token_names() + [
            ("short_description", True),
            ("product_id", False),
            ("supplier_code", False),
            ("sku", False),
            ("upc", False),
            ("ean", False),
            ("brand.name", True),
            ("season.name", True),
            ("characteristics", False),
            ("features", False),
            ("colors.name", True),
            ("categories.name", True),
            ("collections.name", True),
            ("compositions.name", True)
        ]

    @classmethod
    def from_omni(
        cls,
        merchandise,
        inventory_line = None,
        inventory_lines = None,
        gender = "Both",
        currency = "EUR",
        patch = True,
        force = False
    ):
        from . import brand
        from . import color
        from . import season
        from . import category
        from . import collection

        # gathers the complete set of element from the Omni
        # related merchandise element (product or sub-product)
        object_id = merchandise["object_id"]
        modify_date = merchandise["modify_date"]
        company_product_code = merchandise["company_product_code"]
        upc = merchandise["upc"]
        ean = merchandise["ean"]
        weight = merchandise["weight"]
        metadata = merchandise["metadata"] or dict()
        price_compare = metadata.get("compare_price") or None
        discount = metadata.get("discount") or None
        _color = metadata.get("material") or []
        _category = metadata.get("category") or []
        _collection = metadata.get("collection") or []
        _brand = metadata.get("brand")
        _season = metadata.get("season")
        gender = metadata.get("gender") or gender
        order = metadata.get("order")
        discountable = metadata.get("discountable", True)
        orderable = metadata.get("orderable", False)
        sku_field = metadata.get("sku_field")

        # verifies if an inventory line has been provided, if that's the case
        # it's possible to determine a proper modification date for the product
        # taking into account also the modification date of its inventory line
        if inventory_line:
            modify_date_line = inventory_line["modify_date"]
            if modify_date_line > modify_date: modify_date = modify_date_line

        # creates the stocks list in case there are valid inventory lines being
        # passed on the current product update/creation
        stocks = None if inventory_lines == None else []

        # iterates over the complete set of available inventory lines to build the
        # associated stock dictionary with the information on the stock point, this
        # is going to be added to the list of stocks to the product
        for inventory_line in inventory_lines if inventory_lines else []:
            stock_on_hand = inventory_line.get("stock_on_hand", 0)
            stock_reserved = inventory_line.get("stock_reserved", 0)
            stock_in_transit = inventory_line.get("stock_in_transit", 0)
            retail_price = inventory_line.get("retail_price", {}).get("value", 0.0)
            functional_unit = inventory_line.get("functional_unit", None)

            is_valid = functional_unit and functional_unit.get("status") == 1
            if not is_valid: continue

            stock_m = dict(
                store_id = functional_unit["object_id"],
                store_name = functional_unit["name"],
                stock_on_hand = stock_on_hand,
                stock_reserved = stock_reserved,
                stock_in_transit = stock_in_transit,
                retail_price = retail_price
            )
            stocks.append(stock_m)

        colors = _color if isinstance(_color, list) else [_color]
        categories = _category if isinstance(_category, list) else [_category]
        collections = _collection if isinstance(_collection, list) else [_collection]
        colors = [color.Color.ensure_s(_color) for _color in colors]
        categories = [category.Category.ensure_s(_category) for _category in categories]
        collections = [collection.Collection.ensure_s(_collection) for _collection in collections]
        if _brand: _brand = brand.Brand.ensure_s(_brand)
        if _season: _season = season.Season.ensure_s(_season)
        product = cls.get(product_id = company_product_code, raise_e = False)
        if not product: product = cls()

        # in case the weight contains the special 1.0 value then
        # it must be set to invalid as this is considered to be
        # a dummy (and invalid value), this is only performed in
        # case there's a valid season value (eg: watches) set and
        # the patch mode is set (to enable automatic data correction)
        # this is considered a hack to reverse the invalid data source
        # values and should be used with proper care
        if patch and weight == 1.0 and _season: weight = None

        product.product_id = company_product_code
        product.supplier_code = upc
        product.sku = merchandise.get(sku_field) or upc or company_product_code
        product.upc = upc
        product.ean = ean
        product.weight = weight
        product.short_description = merchandise["name"] or company_product_code
        product.description = merchandise["description"]
        product.gender = gender
        product.price_compare = price_compare
        product.currency = currency
        product.order = order
        product.discountable = discountable
        product.orderable = orderable
        product.characteristics = metadata.get("characteristics", [])
        product.features = metadata.get("features", {})
        product.colors = colors
        product.categories = categories
        product.collections = collections
        product.brand = _brand
        product.season = _season

        meta = dict(
            object_id = object_id,
            company_product_code = company_product_code,
            modify_date = modify_date,
            discount = discount
        )
        if hasattr(product, "meta") and product.meta: product.meta.update(meta)
        else: product.meta = meta
        if not stocks == None: product.meta["stocks"] = stocks

        if "stock_on_hand" in merchandise or force:
            product.quantity_hand = merchandise.get("stock_on_hand", 0.0)
        if "retail_price" in merchandise or force:
            # "grabs" the retail price from the original merchandise entity
            # from Omni to be used as the base calculus
            retail_price = merchandise.get("retail_price", 0.0)

            # stores the "original" retail price in the product's metadata
            # storage may be needed latter for update operations
            product.meta["retail_price"] = retail_price
        if "price" in merchandise or force:
            # "grabs" the (untaxed) price from the original merchandise entity
            # from Omni to be used as the base calculus
            untaxed_price = merchandise.get("price", 0.0)

            # stores the "original" untaxed price in the product's metadata
            # storage may be needed latter for update operations
            product.meta["untaxed_price"] = untaxed_price

        # in case all of the required "original" financial information (prices)
        # is available then the price, taxes and price compare are calculated
        if "retail_price" in product.meta and "untaxed_price" in product.meta:
            untaxed_price = _currency.Currency.round(
                product.meta["untaxed_price"] * ((100.0 - discount) / 100.0),
                currency
            ) if discount else product.meta["untaxed_price"]
            product.price = _currency.Currency.round(
                product.meta["retail_price"] * ((100.0 - discount) / 100.0),
                currency
            ) if discount else product.meta["retail_price"]
            product.taxes = product.price - untaxed_price
            if not product.price_compare and discount:
                product.price_compare = product.meta["retail_price"]

        # returns the "final" product instance to the caller so that it's possible
        # to properly save the newly generated product instance according to omni
        return product

    @classmethod
    @appier.operation(
        name = "Import Omni",
        parameters = (
            ("Product", "product", "longtext"),
        ),
        factory = True
    )
    def import_omni_s(cls, product, safe = True):
        product = json.loads(product)
        product = cls.from_omni(product)
        product.save()
        return product

    @classmethod
    @appier.operation(
        name = "Import CSV",
        parameters = (
            ("CSV File", "file", "file"),
            ("Empty source", "empty", bool, False)
        )
    )
    def import_csv_s(cls, file, empty):

        def callback(line):
            from . import color
            from . import brand
            from . import season
            from . import category
            from . import collection
            from . import composition
            from . import measurement

            description,\
            short_description,\
            product_id,\
            gender,\
            price,\
            order,\
            tag,\
            tag_description,\
            farfetch_url,\
            farfetch_male_url,\
            farfetch_female_url,\
            colors,\
            categories,\
            collections, \
            variants,\
            _brand,\
            _season,\
            measurements,\
            compositions, \
            price_provider, \
            price_url = line

            product_id = product_id or None
            price = float(price) if price else None
            order = int(order) if order else None
            tag = tag or None
            tag_description = tag_description or None
            farfetch_url = farfetch_url or None
            farfetch_male_url = farfetch_male_url or None
            farfetch_female_url = farfetch_female_url or None
            price_provider = price_provider or None
            price_url = price_url or None

            colors = colors.split(";") if colors else []
            colors = color.Color.find(name = {"$in" : colors})

            categories = categories.split(";") if categories else []
            categories = category.Category.find(name = {"$in" : categories})

            collections = collections.split(";") if collections else []
            collections = collection.Collection.find(name = {"$in" : collections})

            variants = variants.split(";") if variants else []
            variants = Product.find(product_id = {"$in" : variants})

            _brand = brand.Brand.find(name = _brand) if _brand else None
            _season = season.Season.find(name = _season) if _season else None

            measurements = measurements.split(";") if measurements else []
            measurements = measurement.Measurement.find(name = {"$in" : measurements})

            compositions = compositions.split(";") if compositions else []
            compositions = composition.Composition.find(name = {"$in" : compositions})

            product = cls(
                description = description,
                short_description = short_description,
                product_id = product_id,
                gender = gender,
                price = price,
                order = order,
                tag = tag,
                tag_description = tag_description,
                farfetch_url = farfetch_url,
                farfetch_male_url = farfetch_male_url,
                farfetch_female_url = farfetch_female_url,
                colors = colors,
                categories = categories,
                collections = collections,
                variants = variants,
                brand = _brand,
                season = _season,
                measurements = measurements,
                compositions = compositions,
                price_provider = price_provider,
                price_url = price_url
            )
            product.save()

        if empty: cls.delete_c()
        cls._csv_import(file, callback)

    @classmethod
    @appier.link(name = "Export Simple")
    def simple_csv_url(cls, absolute = False):
        return appier.get_app().url_for(
            "product_api.simple_csv",
            absolute = absolute
        )

    @classmethod
    def _build(cls, model, map):
        super(Product, cls)._build(model, map)

        # retrieves the multiple values from the product model that are
        # going to be used in the calculated attributes construction
        price = model.get("price", None)
        price_compare = model.get("price_compare", None)
        labels = model.get("labels", ())

        # verifies if the current product is considered to be a discounted
        # one, that happens when the price compare is greater than price
        is_discounted = bool(price and price_compare and price_compare > price)

        # calculates the shipping costs taking into account if the price
        # is currently defined for the product defaulting to none otherwise
        if price == None: shipping_cost = None
        else: shipping_cost = bundle.Bundle.eval_shipping(price, 0.0, 1.0, None)

        # sets the multiple attributes of the product that describe it through
        # calculated attributes (as expected by model retriever)
        model["is_discounted"] = is_discounted
        model["is_discounted_s"] = "discounted" if is_discounted else "not-discounted"
        model["shipping_cost"] = shipping_cost
        for label in labels:
            model[label] = label
            model[label + "_s"] = "true"

    def pre_validate(self):
        base.BudyBase.pre_validate(self)
        self.build_images()
        self.build_names()
        self.build_labels()

    def pre_save(self):
        base.BudyBase.pre_save(self)
        if not self.measurements: return
        quantities_hand = [measurement.quantity_hand or 0.0 for measurement in\
            self.measurements if hasattr(measurement, "quantity_hand") and\
            not measurement.quantity_hand == None]
        prices = [measurement.price or 0.0 for measurement in\
            self.measurements if hasattr(measurement, "price") and\
            not measurement.price == None]
        taxes = [measurement.taxes or 0.0 for measurement in\
            self.measurements if hasattr(measurement, "taxes") and\
            not measurement.taxes == None]
        prices_compare = [measurement.price_compare or 0.0 for measurement in\
            self.measurements if hasattr(measurement, "price_compare") and\
            not measurement.price_compare == None]
        self.quantity_hand = sum(quantities_hand) if quantities_hand else None
        self.price = max(prices) if prices else 0.0
        self.taxes = max(taxes) if taxes else 0.0
        self.price_compare = max(prices_compare) if prices_compare else None

        # in case the product is orderable then the quantity on hand must
        # be forced to be greater than zero so that the product can be
        # listed as available (otherwise it would not be listed), this is
        # considered to be a "hack"
        if self.orderable and self.quantity_hand <= 0:
            self.quantity_hand = 1

    def build_images(self):
        thumbnail = self.get_image(size = "thumbnail", order = 1)
        thumbnail = thumbnail or self.get_image(size = "thumbnail")
        image = self.get_image(size = "large", order = 1)
        image = image or self.get_image(size = "large")
        self.thumbnail_url = thumbnail.get_url() if thumbnail else None
        self.image_url = image.get_url() if image else None

    def build_names(self):
        self.brand_s = self.brand.name if self.brand else None
        self.season_s = self.season.name if self.season else None
        self.color_s = self.colors[0].name if self.colors else None
        self.category_s = self.categories[0].name if self.categories else None
        self.collection_s = self.collections[0].name if self.collections else None

    def build_labels(self):
        self._reset_labels()
        self._build_labels(self.brand)
        self._build_labels(self.season)
        self._build_labels(self.colors)
        self._build_labels(self.categories)
        self._build_labels(self.collections)

    def related(self, limit = 6, available = True, enabled = True):
        cls = self.__class__
        kwargs = dict()
        if available: kwargs["quantity_hand"] = {"$gt" : 0}
        if self.collections: kwargs["collections"] = {"$in" : [self.collections[0].id]}
        elif self.categories: kwargs["categories"] = {"$in" : [self.categories[0].id]}
        elif self.colors: kwargs["colors"] = {"$in" : [self.colors[0].id]}
        elif self.brand: kwargs["brand"] = {"$in" : [self.brand.id]}
        elif self.season: kwargs["season"] = {"$in" : [self.season.id]}
        kwargs["id"] = {"$nin" : [self.id]}
        kwargs["sort"] = [("id", 1)]
        count = cls.count(**kwargs)
        skip = self._get_offset(count, limit, kwargs = kwargs)
        delta = skip + limit - count
        if delta > 0: skip = count - skip - delta
        if skip < 0: skip = 0
        find = cls.find_e if enabled else cls.find
        products = find(
            eager = ("images",),
            skip = skip,
            limit = limit,
            map = True,
            **kwargs
        )
        return products

    def get_measurement(self, value, name = None):
        for measurement in self.measurements:
            if not measurement: continue
            if not hasattr(measurement, "value"): continue
            if not hasattr(measurement, "name"): continue
            if not measurement.value == value: continue
            if not measurement.name == name: continue
            return measurement
        return None

    def get_price(
        self,
        currency = None,
        country = None,
        attributes = None
    ):
        if not self.price_provider: return self.price
        method = getattr(self, "get_price_%s" % self.price_provider)
        return method(
            currency = currency,
            country = country,
            attributes = attributes
        )

    def get_taxes(
        self,
        currency = None,
        country = None,
        attributes = None
    ):
        if not self.price_provider: return self.taxes
        method = getattr(self, "get_taxes_%s" % self.price_provider, None)
        if not method: return self.taxes
        return method(
            currency = currency,
            country = country,
            attributes = attributes
        )

    def get_price_ripe(
        self,
        currency = None,
        country = None,
        attributes = None
    ):
        if not self.price_url: return self.price

        result = self.get_availability_ripe(
            currency = currency,
            country = country,
            attributes = attributes
        )
        total = result["total"]
        return total["price_final"]

    def get_taxes_ripe(
        self,
        currency = None,
        country = None,
        attributes = None
    ):
        if not self.price_url: return self.price

        result = self.get_availability_ripe(
            currency = currency,
            country = country,
            attributes = attributes
        )
        total = result["total"]
        return total["ddp"] + total["vat"]

    def get_availability_ripe(
        self,
        currency = None,
        country = None,
        attributes = None
    ):
        attributes_m = json.loads(attributes)
        p = []
        parts = attributes_m.get("parts", {})
        embossing = attributes_m.get("embossing", None)
        letters = attributes_m.get("letters", None)

        for key, value in appier.legacy.iteritems(parts):
            material = value["material"]
            color = value["color"]
            triplet = "%s:%s:%s" % (key, material, color)
            p.append(triplet)

        params = dict(
            product_id = self.product_id,
            p = p
        )
        if currency: params["currency"] = currency
        if country: params["country"] = country
        if embossing: params["embossing"] = embossing
        if letters: params["letters"] = letters

        result = appier.get(
            self.price_url,
            params = params
        )
        return result

    def get_currency(self, currency = None):
        if not self.price_provider: return self.currency or currency
        method = getattr(self, "get_currency_%s" % self.price_provider)
        return method(currency = currency)

    def get_currency_ripe(self, currency = None):
        return self.currency or currency

    def get_image(self, size = None, order = None):
        for image in self.images:
            if not image: continue
            if not hasattr(image, "size"): continue
            if not hasattr(image, "order"): continue
            is_size = size == None or image.size == size
            if not is_size: continue
            is_order = order == None or image.order == order
            if not is_order: continue
            return image
        return None

    def get_size(self, currency = None, country = None, attributes = None):
        if not self.price_provider: return None, None
        method = getattr(self, "get_size_%s" % self.price_provider)
        return method(
            country = country,
            attributes = attributes
        )

    def get_size_ripe(
        self,
        currency = None,
        country = None,
        attributes = None
    ):
        attributes_m = json.loads(attributes)
        size = attributes_m["size"]
        scale = attributes_m["scale"]
        gender = attributes_m["gender"]

        if gender == "male": converter = lambda native: ((native - 17) / 2) + 36
        else: converter = lambda native: ((native - 17) / 2) + 34

        return converter(size), scale

    @appier.operation(
        name = "Add Collection",
        parameters = (
            (
                "Collection",
                "collection",
                appier.reference("Collection", name = "id")
            ),
        )
    )
    def add_collection_s(self, collection):
        if not collection: return
        if collection in self.collections: return
        self.collections.append(collection)
        self.save()

    @appier.operation(
        name = "Remove Collection",
        parameters = (
            (
                "Collection",
                "collection",
                appier.reference("Collection", name = "id")
            ),
        )
    )
    def remove_collection_s(self, collection):
        if not collection: return
        if not collection in self.collections: return
        self.collections.remove(collection)
        self.save()

    @appier.operation(
        name = "Add Image",
        parameters = (
            (
                "Image",
                "image",
                appier.reference("Media", name = "id")
            ),
        )
    )
    def add_image_s(self, image):
        if not image: return
        if image in self.images: return
        self.images.append(image)
        self.save()

    @appier.operation(
        name = "Remove Image",
        parameters = (
            (
                "Image",
                "image",
                appier.reference("Media", name = "id")
            ),
        )
    )
    def remove_image_s(self, image):
        if not image: return
        if not image in self.images: return
        self.images.remove(image)
        self.save()

    @appier.operation(name = "Fix")
    def fix_s(self):
        if not self.exists(): return
        self.save()

    @appier.operation(name = "Ensure Quantity", level = 2, devel = True)
    def ensure_quantity_s(self, quantity = 1.0):
        if self.quantity_hand: return
        self.quantity_hand = quantity
        self.save()

    @appier.operation(name = "Ensure Price", level = 2, devel = True)
    def ensure_price_s(self, price = 100.0):
        if self.price: return
        self.price = price
        self.save()

    @appier.operation(
        name = "Notify",
        parameters = (("Email", "email", str),)
    )
    def notify(self, name = None, *args, **kwargs):
        name = name or "product.new"
        product = self.reload(map = True)
        receiver = kwargs.get("email", None)
        appier_extras.admin.Event.notify_g(
            name,
            arguments = dict(
                params = dict(
                    payload = product,
                    product = product,
                    receiver = receiver,
                    extra = kwargs
                )
            )
        )

    @appier.operation(
        name = "Share",
        parameters = (
            ("Sender", "sender", appier.legacy.UNICODE),
            ("Email", "email", str)
        )
    )
    def share(self, *args, **kwargs):
        self.notify(name = "product.share", *args, **kwargs)

    @appier.operation(
        name = "Quote",
        parameters = (
            ("Requester", "requester", appier.legacy.UNICODE),
            ("Email", "email", str),
            ("Phone", "phone", str)
        )
    )
    def quote(self, *args, **kwargs):
        # patches the email attribute for the quote call so
        # that the receiver value can be "manipulated" by
        # configuration as it must be a admin user of the
        # store by default and not the end-user
        kwargs_quote = dict(kwargs)
        kwargs_quote["_email"] = kwargs_quote.pop("email", None)

        # runs the notification process for the product quote making
        # sure that all of the notification names are used
        self.notify(name = "product.quote", *args, **kwargs_quote)
        self.notify(name = "product.quote.confirmation", *args, **kwargs)

    @appier.view(name = "Measurements")
    def measurements_v(self, *args, **kwargs):
        kwargs["sort"] = kwargs.get("sort", [("id", 1)])
        return appier.lazy_dict(
            model = self.measurements._target,
            kwargs = kwargs,
            entities = appier.lazy(lambda: self.measurements.find(*args, **kwargs)),
            page = appier.lazy(lambda: self.measurements.paginate(*args, **kwargs))
        )

    @property
    def quantity(self):
        return self.quantity_hand

    @property
    def discount(self):
        if not self.price: return commons.Decimal(0.0)
        if not self.price_compare: return commons.Decimal(0.0)
        return self.price_compare - self.price

    @property
    def discount_percent(self):
        if not self.discount: return commons.Decimal(0.0)
        return self.discount / self.price_compare * commons.Decimal(100.0)

    @property
    def is_parent(self):
        if not hasattr(self, "measurements"): return False
        if not self.measurements: return False
        return len(self.measurements) > 0

    @property
    def is_discounted(self):
        return self.discount > 0.0

    @property
    def is_discountable(self):
        return self.discountable

    @property
    def is_price_provided(self):
        return True if self.price_provider else False

    def _reset_labels(self):
        self.labels = []

    def _build_labels(self, groups):
        if not isinstance(groups, (list, tuple, appier.References)):
            groups = (groups,) if groups else []
        for group in groups:
            for label in group.labels:
                if label in self.labels: continue
                self.labels.append(label)

    def _get_offset(self, count, limit, kwargs):
        return self._get_offset_offset(count, limit, kwargs)

    def _get_offset_simple(self, count, limit, kwargs):
        return self.id % count

    def _get_offset_offset(self, count, limit, kwargs):
        cls = self.__class__
        kwargs = dict(kwargs)
        kwargs["id"] = {"$lt" : self.id}
        offset = cls.count(**kwargs)
        return offset



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from . import base

class Referral(base.BudyBase):

    name = appier.field(
        index = True,
        default = True
    )

    @classmethod
    def validate(cls):
        return super(Referral, cls).validate() + [
            appier.not_null("name"),
            appier.not_empty("name"),
            appier.string_gt("name", 3),
            appier.string_lt("name", 64),
            appier.not_duplicate("name", cls._name())
        ]

    @classmethod
    def list_names(cls):
        return ["id", "name"]

    @classmethod
    def order_name(cls):
        return ["id", -1]

    @classmethod
    @appier.operation(
        name = "Import CSV",
        parameters = (
            ("CSV File", "file", "file"),
            ("Empty source", "empty", bool, False)
        )
    )
    def import_csv_s(cls, file, empty):
        def callback(line):
            store, name = line
            composed_name = "%s|%s" % (name, store)
            referral = cls(
                name = composed_name
            )
            referral.save()

        if empty: cls.delete_c()
        cls._csv_import(file, callback)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import commons
import logging
import unittest

import appier

import budy

class OrderTest(unittest.TestCase):

    def setUp(self):
        self.app = budy.BudyApp(level = logging.ERROR)

    def tearDown(self):
        self.app.unload()
        adapter = appier.get_adapter()
        adapter.drop_db()

    def test_basic(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        self.assertEqual(product.short_description, "product")
        self.assertEqual(product.gender, "Male")
        self.assertEqual(product.price, 10.0)

        order = budy.Order()
        order.save()

        self.assertEqual(type(order.key), appier.legacy.UNICODE)
        self.assertEqual(type(order.total), commons.Decimal)
        self.assertEqual(len(order.lines), 0)
        self.assertEqual(order.id, 1)
        self.assertEqual(order.reference, "BD-000001")
        self.assertEqual(order.currency, None)
        self.assertEqual(order.total >= 0.0, True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.date, None)
        self.assertEqual(order.notifications, [])

        self.assertRaises(appier.AssertionError, order.verify_base)
        self.assertRaises(appier.AssertionError, order.verify_shippable)
        self.assertRaises(appier.AssertionError, order.mark_paid_s)

        order = order.reload()

        self.assertEqual(type(order.key), appier.legacy.UNICODE)
        self.assertEqual(type(order.total), commons.Decimal)
        self.assertEqual(len(order.lines), 0)
        self.assertEqual(len(order.reference), 9)
        self.assertEqual(order.id, 1)
        self.assertEqual(order.reference, "BD-000001")
        self.assertEqual(order.currency, None)
        self.assertEqual(order.total >= 0.0, True)
        self.assertEqual(order.status, "created")
        self.assertEqual(order.paid, False)
        self.assertEqual(order.date, None)
        self.assertEqual(order.notifications, [])
        self.assertEqual(order.reference.startswith("BD-"), True)

        self.assertRaises(appier.AssertionError, order.verify_base)
        self.assertRaises(appier.AssertionError, order.verify_shippable)
        self.assertRaises(appier.AssertionError, order.mark_paid_s)

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)
        order.verify_base()

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        self.assertRaises(appier.AssertionError, order.verify_shippable)
        self.assertRaises(appier.AssertionError, order.mark_paid_s)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        order.mark_waiting_payment_s()

        self.assertEqual(order.status, "waiting_payment")
        self.assertEqual(order.paid, False)

        order.mark_paid_s()

        self.assertEqual(order.status, "paid")
        self.assertEqual(order.paid, True)

        self.assertRaises(appier.AssertionError, order.mark_paid_s)

    def test_referral(self):
        order = budy.Order()
        order.save()

        referral = budy.Referral(name = "name")
        referral.save()

        order.set_referral_s(referral)

        self.assertEqual(len(order.referrals), 1)
        self.assertEqual(order.referrals[0].name, "name")

    def test_store(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.currency, None)
        self.assertEqual(order.payment_currency, None)

        order.refresh_s(
            currency = order.payment_currency or order.currency,
            country = order.shipping_country or order.country
        )

        self.assertEqual(order.currency, None)
        self.assertEqual(order.payment_currency, None)

        store = budy.Store(name = "store", currency_code = "GBP")
        store.save()

        account = budy.BudyAccount(
            username = "account",
            email = "account@account.com",
            password = "password",
            password_confirm = "password",
            store = store
        )
        account.save()

        order.set_account_s(account)

        self.assertEqual(order.account.username, "account")
        self.assertEqual(order.store.name, "store")
        self.assertEqual(order.currency, None)
        self.assertEqual(order.payment_currency, "GBP")

        order.refresh_s(
            currency = order.payment_currency or order.currency,
            country = order.shipping_country or order.country
        )

        self.assertEqual(order.currency, "GBP")
        self.assertEqual(order.payment_currency, "GBP")

        self.assertRaises(appier.AssertionError, order.verify_store)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()
        store.address = address
        store.save()

        order = order.reload()

        order.verify_store()

    def test_voucher(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        voucher = budy.Voucher(amount = 5.0)
        voucher.save()

        order.add_voucher_s(voucher)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 5.0)
        self.assertEqual(order.total, 15.0)
        self.assertEqual(order.payable, 15.0)
        self.assertEqual(order.discountable, 20.0)
        self.assertEqual(isinstance(order.sub_total, commons.Decimal), True)
        self.assertEqual(isinstance(order.discount, commons.Decimal), True)
        self.assertEqual(isinstance(order.total, commons.Decimal), True)
        self.assertEqual(isinstance(order.payable, commons.Decimal), True)
        self.assertEqual(isinstance(order.discountable, commons.Decimal), True)

        order.use_vouchers_s()
        order.verify_vouchers()
        voucher = voucher.reload()

        self.assertEqual(voucher.is_valid(), False)
        self.assertEqual(voucher.used_amount, 5.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

        small_voucher = budy.Voucher(amount = 1.0)
        small_voucher.save()

        order.set_voucher_s(small_voucher)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 1.0)
        self.assertEqual(order.total, 19.0)
        self.assertEqual(order.payable, 19.0)
        self.assertEqual(order.discountable, 20.0)
        self.assertEqual(isinstance(order.sub_total, commons.Decimal), True)
        self.assertEqual(isinstance(order.discount, commons.Decimal), True)
        self.assertEqual(isinstance(order.total, commons.Decimal), True)
        self.assertEqual(isinstance(order.payable, commons.Decimal), True)

        large_voucher = budy.Voucher(amount = 100.0)
        large_voucher.save()

        order.set_voucher_s(large_voucher)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.discountable, 20.0)

        order.use_vouchers_s()
        order.verify_vouchers()
        voucher = large_voucher.reload()

        self.assertEqual(voucher.is_valid(), True)
        self.assertEqual(voucher.used_amount, 20.0)
        self.assertEqual(voucher.open_amount, 80.0)
        self.assertEqual(voucher.usage_count, 1)

        order.unmark_paid_s()

        percent_voucher = budy.Voucher(percentage = 10.0)
        percent_voucher.save()

        order.set_voucher_s(percent_voucher)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 2.0)
        self.assertEqual(order.total, 18.0)
        self.assertEqual(order.payable, 18.0)
        self.assertEqual(order.discountable, 20.0)

        order.use_vouchers_s()
        order.verify_vouchers()
        voucher = percent_voucher.reload()

        self.assertEqual(voucher.is_valid(), True)
        self.assertEqual(voucher.percentage, 10.0)
        self.assertEqual(voucher.used_amount, 0.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

    def test_duplicate_pay(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        def _pay_dummy(payment_data):
            order.payment_data = dict(method = "dummy")
            return False

        order.pay_s(
            payment_data = dict(type = "simple"),
            payment_function = _pay_dummy
        )

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.status, "waiting_payment")

        self.assertRaises(
            appier.SecurityError,
            lambda: order.pay_s(
                payment_data = dict(type = "simple"),
                payment_function = _pay_dummy
            )
        )

    def test_voucher_use(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        voucher = budy.Voucher(amount = 1.0)
        voucher.save()

        order.set_voucher_s(voucher)

        order.pay_s(
            payment_data = dict(type = "simple"),
            strict = False
        )

        self.assertEqual(voucher.is_valid(), False)
        self.assertEqual(voucher.amount, 1.0)
        self.assertEqual(voucher.used_amount, 1.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.status, "waiting_payment")

        order.end_pay_s(
            payment_data = dict(type = "simple"),
            strict = False
        )

        self.assertEqual(voucher.is_valid(), False)
        self.assertEqual(voucher.amount, 1.0)
        self.assertEqual(voucher.used_amount, 1.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.status, "paid")

        order.date = None
        order.status = "created"
        order.unmark_paid_s()

        voucher = budy.Voucher(amount = 1.0)
        voucher.save()

        order.set_voucher_s(voucher)

        order.pay_s(
            payment_data = dict(type = "simple"),
            strict = False
        )

        self.assertEqual(voucher.is_valid(), False)
        self.assertEqual(voucher.amount, 1.0)
        self.assertEqual(voucher.used_amount, 1.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.status, "waiting_payment")

        order.cancel_s()

        voucher = voucher.reload()

        self.assertEqual(voucher.is_valid(), True)
        self.assertEqual(voucher.amount, 1.0)
        self.assertEqual(voucher.used_amount, 0.0)
        self.assertEqual(voucher.open_amount, 1.0)
        self.assertEqual(voucher.usage_count, 0)

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.status, "canceled")

        order.date = None
        order.status = "created"
        order.unmark_paid_s()

        voucher = budy.Voucher(amount = 1.0)
        voucher.save()

        order.set_voucher_s(voucher)

        order.pay_s(
            payment_data = dict(type = "simple"),
            strict = False
        )

        order.end_pay_s(
            payment_data = dict(type = "simple"),
            strict = False
        )

        self.assertEqual(voucher.is_valid(), False)
        self.assertEqual(voucher.amount, 1.0)
        self.assertEqual(voucher.used_amount, 1.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.status, "paid")

        order.cancel_s()

        voucher = voucher.reload()

        self.assertEqual(voucher.is_valid(), False)
        self.assertEqual(voucher.amount, 1.0)
        self.assertEqual(voucher.used_amount, 1.0)
        self.assertEqual(voucher.open_amount, 0.0)
        self.assertEqual(voucher.usage_count, 1)

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.status, "canceled")

    def test_discount(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(order.payable, 20.0)
        self.assertEqual(order.discountable, 20.0)

        order.discount_fixed = 20.0
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.discountable, 20.0)

        voucher = budy.Voucher(amount = 5.0)
        voucher.save()

        order.discount_fixed = 18.0
        order.set_voucher_s(voucher)
        order.use_vouchers_s()
        order.verify_vouchers()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.discountable, 20.0)

        voucher = voucher.reload()

        self.assertEqual(voucher.is_valid(), True)
        self.assertEqual(voucher.used_amount, 2.0)
        self.assertEqual(voucher.open_amount, 3.0)
        self.assertEqual(voucher.usage_count, 1)

    def test_discountable_full(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order(
            discountable_full = True,
            shipping_fixed = 10.0
        )
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 30.0)
        self.assertEqual(order.payable, 30.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 30.0)

        order.discount_fixed = 20.0
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 10.0)
        self.assertEqual(order.payable, 10.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 30.0)

        order.discount_fixed = 30.0
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 30.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 30.0)

        order.discountable_full = False
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 10.0)
        self.assertEqual(order.payable, 10.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 20.0)

        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            price_compare = 12.0
        )
        product.save()

        order = budy.Order(
            discountable_full = True,
            shipping_fixed = 10.0
        )
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 30.0)
        self.assertEqual(order.payable, 30.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 30.0)

        order.discount_fixed = 20.0
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 10.0)
        self.assertEqual(order.payable, 10.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 30.0)

        order.discount_fixed = 30.0
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 30.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 30.0)

        order.discountable_full = False
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 30.0)
        self.assertEqual(order.payable, 30.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 0.0)

    def test_non_discountable(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            discountable = False
        )
        product.save()

        order = budy.Order(
            shipping_fixed = 10.0
        )
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 30.0)
        self.assertEqual(order.payable, 30.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 0.0)

        order.discount_fixed = 20.0
        order.save()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 30.0)
        self.assertEqual(order.payable, 30.0)
        self.assertEqual(order.shipping_cost, 10.0)
        self.assertEqual(order.discountable, 0.0)

    def test_taxes(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            taxes = 3.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(order.taxes, 6.0)
        self.assertEqual(order.payable, 20.0)
        self.assertEqual(order.discountable, 20.0)

    def test_quantity(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            quantity_hand = 1.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product

        self.assertRaises(appier.AssertionError, order_line.save)

        product.quantity_hand = 2.0
        product.save()

        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.is_open(), True)
        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)

        product.quantity_hand = 1.0
        product.save()

        self.assertEqual(order.is_open(), True)
        self.assertEqual(order.is_valid(), False)
        self.assertEqual(order_line.is_valid_quantity(), False)
        self.assertRaises(appier.AssertionError, order.save)
        self.assertRaises(appier.AssertionError, order.mark_waiting_payment_s)

        product.quantity_hand = None
        product.save()

        self.assertEqual(order.is_open(), True)
        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)

        product.quantity_hand = 1.0
        product.save()

        order.paid = True

        self.assertEqual(order.is_open(), False)
        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), False)

        order.paid = False

        product.quantity_hand = 2.0
        product.save()

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        order.save()
        order.mark_waiting_payment_s()
        order.mark_paid_s()

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.status, "paid")

        product.quantity_hand = 0.0
        product.save()

        order.save()

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.status, "paid")

    def test_measurements(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            quantity_hand = None
        )
        product.save()

        measurement = budy.Measurement(
            name = "size",
            value = 12,
            value_s = "12",
            price = 12.0,
            quantity_hand = None,
            product = product
        )
        measurement.save()

        product.measurements.append(measurement)
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product

        self.assertRaises(appier.AssertionError, order_line.save)
        self.assertEqual(order_line.is_valid(), False)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order_line.is_valid_price(), True)
        self.assertEqual(order_line.is_valid_size(), False)

        order_line = budy.OrderLine(
            price = 12.0,
            quantity = 2.0,
            size = 12,
            scale = 1
        )
        order_line.product = product

        self.assertEqual(order_line.is_valid(), True)
        self.assertEqual(order_line.size_s, None)

        order_line.ensure_size_s()

        self.assertEqual(order_line.size_s, "12")

        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order.sub_total, 24.0)
        self.assertEqual(order.discount, 0.0)
        self.assertEqual(order.total, 24.0)
        self.assertEqual(order.payable, 24.0)
        self.assertEqual(order.discountable, 24.0)

    def test_free(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        voucher = budy.Voucher(amount = 20.0)
        voucher.save()

        order.add_voucher_s(voucher)

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.discountable, 20.0)
        self.assertEqual(order.paid, False)
        self.assertEqual(isinstance(order.sub_total, commons.Decimal), True)
        self.assertEqual(isinstance(order.discount, commons.Decimal), True)
        self.assertEqual(isinstance(order.total, commons.Decimal), True)
        self.assertEqual(isinstance(order.payable, commons.Decimal), True)
        self.assertEqual(isinstance(order.discountable, commons.Decimal), True)

        order.pay_s()

        self.assertEqual(order.sub_total, 20.0)
        self.assertEqual(order.discount, 20.0)
        self.assertEqual(order.total, 0.0)
        self.assertEqual(order.payable, 0.0)
        self.assertEqual(order.discountable, 20.0)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.payment_data, {})
        self.assertEqual(isinstance(order.sub_total, commons.Decimal), True)
        self.assertEqual(isinstance(order.discount, commons.Decimal), True)
        self.assertEqual(isinstance(order.total, commons.Decimal), True)
        self.assertEqual(isinstance(order.payable, commons.Decimal), True)
        self.assertEqual(isinstance(order.discountable, commons.Decimal), True)

    def test_closed(self):
        order = budy.Order()
        order.save()

        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        self.assertEqual(order.is_open(), True)
        self.assertEqual(order.is_closed(), False)

        order.mark_waiting_payment_s()

        self.assertEqual(order.is_open(), False)
        self.assertEqual(order.is_closed(), True)

    def test_payment_method(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        address = budy.Address(
            first_name = "first name",
            last_name = "last name",
            address = "address",
            city = "city"
        )
        address.save()

        order.shipping_address = address
        order.billing_address = address
        order.email = "username@email.com"
        order.save()

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.status, "created")

        self.assertRaises(
            appier.SecurityError,
            lambda: order.pay_s(payment_data = dict(type = "simple"))
        )

        order.pay_s(
            payment_data = dict(type = "simple"),
            strict = False
        )

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, False)
        self.assertEqual(order.status, "waiting_payment")

        self.assertRaises(
            appier.SecurityError,
            lambda: order.end_pay_s(
                payment_data = dict(type = "simple"),
                strict = True
            )
        )

        order.end_pay_s(
            payment_data = dict(type = "simple")
        )

        self.assertEqual(order.is_valid(), True)
        self.assertEqual(order_line.is_valid_quantity(), True)
        self.assertEqual(order.paid, True)
        self.assertEqual(order.status, "paid")

    def test__build_notes(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        order = budy.Order()
        order.save()

        order_line = budy.OrderLine(quantity = 2.0)
        order_line.product = product
        order_line.save()
        order.add_line_s(order_line)

        self.assertEqual(order_line.quantity, 2.0)
        self.assertEqual(order_line.total, 20.0)
        self.assertEqual(order.total, 20.0)
        self.assertEqual(len(order.lines), 1)

        notes = order._build_notes()

        self.assertEqual(notes, "Budy order - BD-000001")

        order_line.attributes = json.dumps(dict(initials = "IN"))
        order_line.save()

        order = order.reload()
        notes = order._build_notes()

        self.assertEqual(notes, "Budy order - BD-000001\nOrder line - product | initials = IN")

        order_line.attributes = json.dumps(dict(initials = "IN", engraving = "gold"))
        order_line.save()

        order = order.reload()
        notes = order._build_notes()

        self.assertEqual(notes, "Budy order - BD-000001\nOrder line - product | engraving = gold\nOrder line - product | initials = IN")

        order_line.attributes = "INVALID STRING"
        order_line.save()

        order = order.reload()
        notes = order._build_notes()

        self.assertEqual(notes, "Budy order - BD-000001")

        order_line.attributes = None
        order_line.save()

        order = order.reload()
        notes = order._build_notes()

        self.assertEqual(notes, "Budy order - BD-000001")

        order_line.attributes = 12345
        order_line.save()

        order = order.reload()
        notes = order._build_notes()

        self.assertEqual(notes, "Budy order - BD-000001")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import commons

import appier
import appier_extras

from . import base

class Voucher(base.BudyBase):

    key = appier.field(
        index = True,
        safe = True,
        immutable = True
    )

    amount = appier.field(
        type = commons.Decimal,
        index = True,
        initial = commons.Decimal(0.0),
        safe = True,
        immutable = True
    )

    used_amount = appier.field(
        type = commons.Decimal,
        initial = commons.Decimal(0.0),
        index = True,
        safe = True
    )

    percentage = appier.field(
        type = commons.Decimal,
        initial = commons.Decimal(0.0),
        index = True,
        safe = True,
        immutable = True
    )

    currency = appier.field(
        index = True,
        safe = True,
        observations = """The base currency to be used for
        the voucher, usage of the voucher under other currency
        will be performed using proper exchange rate conversion"""
    )

    start = appier.field(
        type = int,
        index = True,
        safe = True,
        meta = "datetime"
    )

    expiration = appier.field(
        type = int,
        index = True,
        safe = True,
        meta = "datetime"
    )

    usage_count = appier.field(
        type = int,
        initial = 0,
        index = True,
        safe = True,
        observations = """The number of time the voucher has
        already been used, if this value reached the usage limit
        the voucher is marked as used and no further usages are
        possible"""
    )

    usage_limit = appier.field(
        type = int,
        initial = 0,
        index = True,
        safe = True,
        observations = """The maximum number of usages of the
        voucher before it's considered invalid, if the number
        is set to zero the voucher is considered to have no
        limit in the number of usages"""
    )

    unlimited = appier.field(
        type = bool,
        initial = False,
        index = True,
        safe = True,
        observations = """Flag that indicated if the value based
        voucher should not have its used amount deducted and instead
        be considered an unlimited voucher"""
    )

    used = appier.field(
        type = bool,
        initial = False,
        index = True,
        safe = True,
        observations = """Calculated value used to determine if
        the voucher is already used and should not be used anymore
        under any circumstance"""
    )

    @classmethod
    def validate(cls):
        return super(Voucher, cls).validate() + [
            appier.not_empty("key"),
            appier.not_duplicate("key", cls._name()),

            appier.not_null("amount"),
            appier.gte("amount", 0.0),

            appier.not_null("used_amount"),
            appier.gte("used_amount", 0.0),

            appier.not_null("percentage"),
            appier.gte("percentage", 0.0),
            appier.lte("percentage", 100.0),

            appier.not_null("usage_count"),
            appier.gte("usage_count", 0),

            appier.not_null("usage_limit"),
            appier.gte("usage_limit", 0),

            appier.not_null("unlimited"),

            appier.not_null("used")
        ]

    @classmethod
    def list_names(cls):
        return [
            "description",
            "created",
            "amount",
            "percentage",
            "start",
            "expiration",
            "used"
        ]

    @classmethod
    def order_name(cls):
        return ["id", -1]

    @classmethod
    def is_snapshot(cls):
        return True

    @classmethod
    @appier.operation(
        name = "Create Value",
        parameters = (
            ("Key", "key", str),
            ("Amount", "amount", commons.Decimal),
            ("Currency", "currency", str),
            ("Unlimited", "unlimited", bool, False)
        ),
        factory = True
    )
    def create_value_s(cls, key, amount, currency, unlimited):
        voucher = cls(
            key = key,
            amount = amount,
            currency = currency,
            unlimited = unlimited
        )
        voucher.save()
        return voucher

    @classmethod
    @appier.operation(
        name = "Create Value Multiple",
        parameters = (
            ("Key", "key", str),
            ("Amount", "amount", commons.Decimal),
            ("Currency", "currency", str),
            ("Unlimited", "unlimited", bool, False),
            ("Count", "count", int, 1)
        )
    )
    def create_value_multiple_s(cls, key, amount, currency, unlimited, count):
        key = key or cls.secret_g()
        for index in appier.legacy.xrange(count):
            voucher = cls(
                key = key + "-" + str(index),
                amount = amount,
                currency = currency,
                unlimited = unlimited
            )
            voucher.save()

    @classmethod
    @appier.operation(
        name = "Create Percentage",
        parameters = (
            ("Key", "key", str),
            ("Percentage", "percentage", commons.Decimal)
        ),
        factory = True
    )
    def create_percentage_s(cls, key, percentage):
        voucher = cls(
            key = key,
            percentage = percentage
        )
        voucher.save()
        return voucher

    @classmethod
    @appier.operation(
        name = "Create Percentage Multiple",
        parameters = (
            ("Key", "key", str),
            ("Percentage", "percentage", commons.Decimal),
            ("Count", "count", int, 1)
        )
    )
    def create_percentage_multiple_s(cls, key, percentage, count):
        key = key or cls.secret_g()
        for index in appier.legacy.xrange(count):
            voucher = cls(
                key = key + "-" + str(index),
                percentage = percentage
            )
            voucher.save()

    def pre_create(self):
        base.BudyBase.pre_create(self)
        if not hasattr(self, "key") or not self.key:
            self.key = self.secret()
        self.description = self.key[:8]

    def pre_update(self):
        base.BudyBase.pre_update(self)
        if not self.used and self.is_used(): self.used = True
        if self.used and not self.is_used(): self.used = False

    def use_s(self, amount, currency = None):
        amount_l = self.to_local(amount, currency)
        appier.verify(self.is_valid(amount = amount, currency = currency))
        if self.is_value and not self.unlimited:
            self.used_amount += commons.Decimal(amount_l)
        self.usage_count += 1
        self.save()

    def disuse_s(self, amount, currency = None):
        appier.verify(self.usage_count > 0)
        amount_l = self.to_local(amount, currency)
        if self.is_value and not self.unlimited:
            self.used_amount -= commons.Decimal(amount_l)
            self.used_amount = max(self.used_amount, commons.Decimal(0.0))
        self.usage_count -= 1
        self.save()

    def discount(self, amount, currency = None):
        if self.is_percent: return self.open_amount_p(amount, currency = currency)
        else: return self.open_amount_r(currency = currency)

    def to_local(self, amount, currency, reversed = False):
        from . import exchange_rate
        if not amount: return amount
        if not currency: return amount
        if not self.currency: return amount
        if currency == self.currency: return amount
        return exchange_rate.ExchangeRate.convert(
            amount,
            currency,
            self.currency,
            reversed = reversed,
            rounder = commons.floor
        )

    def to_remote(self, amount, currency, reversed = True):
        from . import exchange_rate
        if not amount: return amount
        if not currency: return amount
        if not self.currency: return amount
        if currency == self.currency: return amount
        return exchange_rate.ExchangeRate.convert(
            amount,
            self.currency,
            currency,
            reversed = reversed,
            rounder = commons.floor
        )

    def is_used(self):
        if self.usage_limit and self.usage_count >= self.usage_limit: return True
        if self.amount and commons.Decimal(self.used_amount) >= commons.Decimal(self.amount): return True
        return False

    def is_valid(self, amount = None, currency = None):
        current = time.time()
        amount_l = self.to_local(amount, currency)
        if self.is_used(): return False
        if self.used: return False
        if not self.enabled: return False
        if self.start and current < self.start: return False
        if self.expiration and current > self.expiration: return False
        if self.amount and amount_l and commons.Decimal(amount_l) > commons.Decimal(self.open_amount): return False
        if currency and not self.is_valid_currency(currency): return False
        if not self.amount and not self.percentage: return False
        return True

    def is_valid_currency(self, currency):
        from . import exchange_rate
        if not currency: return True
        if not self.currency: return True
        if currency == self.currency: return True
        has_to_remote = exchange_rate.ExchangeRate.has_rate(self.currency, currency)
        has_to_local = exchange_rate.ExchangeRate.has_rate(currency, self.currency)
        if has_to_remote and has_to_local: return True
        return False

    def open_amount_r(self, currency = None):
        open_amount = commons.Decimal(self.amount) - commons.Decimal(self.used_amount)
        return self.to_remote(open_amount, currency)

    def open_amount_p(self, amount, currency = None):
        from . import currency as _currency
        decimal = self.percentage / 100.0
        open_amount = commons.Decimal(amount) * decimal
        if not currency: return open_amount
        return _currency.Currency.round(open_amount, currency)

    @appier.operation(
        name = "Notify",
        parameters = (("Email", "email", str),)
    )
    def notify(self, name = None, *args, **kwargs):
        name = name or "voucher.new"
        voucher = self.reload(map = True)
        receiver = kwargs.get("email", None)
        appier_extras.admin.Event.notify_g(
            name,
            arguments = dict(
                params = dict(
                    payload = voucher,
                    voucher = voucher,
                    receiver = receiver,
                    extra = kwargs
                )
            )
        )

    @appier.operation(
        name = "Remind",
        parameters = (("Email", "email", str),)
    )
    def remind(self, *args, **kwargs):
        self.notify(name = "voucher.remind", *args, **kwargs)

    @appier.view(name = "All Orders")
    def all_orders_v(self, *args, **kwargs):
        from . import order
        kwargs["sort"] = kwargs.get("sort", [("created", -1)])
        kwargs.update(vouchers = {"$in" : (self.id,)})
        return appier.lazy_dict(
            model = order.Order,
            kwargs = kwargs,
            entities = appier.lazy(lambda: order.Order.find(*args, **kwargs)),
            page = appier.lazy(lambda: order.Order.paginate(*args, **kwargs)),
            names = ["reference", "created", "total", "currency", "status"]
        )

    @appier.view(name = "Paid Orders")
    def paid_orders_v(self, *args, **kwargs):
        from . import order
        kwargs["sort"] = kwargs.get("sort", [("created", -1)])
        kwargs.update(
            vouchers = {"$in" : (self.id,)},
            paid = True,
        )
        return appier.lazy_dict(
            model = order.Order,
            kwargs = kwargs,
            entities = appier.lazy(lambda: order.Order.find(*args, **kwargs)),
            page = appier.lazy(lambda: order.Order.paginate(*args, **kwargs)),
            names = ["reference", "created", "total", "currency", "status"]
        )

    @property
    def open_amount(self):
        return self.open_amount_r()

    @property
    def is_percent(self):
        if self.amount: return False
        if self.percentage: return True
        raise appier.OperationalError(
            message = "No amount or percentage defined"
        )

    @property
    def is_value(self):
        return not self.is_percent

    @property
    def is_unlimited(self):
        return self.unlimited



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

import budy

class OrderController(appier.Controller):

    @appier.route("/orders/me", "GET")
    @appier.route("/orders/me/<str:status>", "GET")
    @appier.ensure(token = "user")
    def me(self, status = "waiting_payment"):
        account = budy.BudyAccount.from_session()
        orders = account.get_store_orders(status = status)
        return self.template(
            "order/me.html.tpl",
            orders = orders
        )

    @appier.route("/orders/<str:key>/mark_paid", "GET")
    @appier.ensure(token = "user")
    def mark_paid(self, key):
        order = budy.Order.get(key = key)
        account = budy.BudyAccount.from_session()
        order.verify_account(account)
        order.mark_paid_s()
        order.notify_s()
        return self.redirect(
            self.url_for("order.me")
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier
import appier_extras

from . import bag
from . import order

class BudyAccount(appier_extras.admin.Account):

    PREFIXES = appier_extras.admin.Account.PREFIXES + [
        "budy."
    ]

    GENDER_S = dict(
        Male = "Male",
        Female = "Female"
    )

    first_name = appier.field(
        index = True
    )

    last_name = appier.field(
        index = True
    )

    gender = appier.field(
        meta = "enum",
        enum = GENDER_S
    )

    birth_date = appier.field(
        type = int,
        index = True,
        meta = "date"
    )

    country = appier.field(
        meta = "country"
    )

    phone_number = appier.field()

    receive_newsletters = appier.field(
        type = bool,
        initial = False
    )

    store = appier.field(
        type = appier.reference(
            "Store",
            name = "id"
        ),
        eager = True
    )

    addresses = appier.field(
        type = appier.references(
            "Address",
            name = "id"
        )
    )

    @classmethod
    def _build(cls, model, map):
        super(BudyAccount, cls)._build(model, map)
        id = model.get("id", None)
        first_name = model.get("first_name", None)
        last_name = model.get("last_name", None)
        full_name = cls._build_full_name(first_name, last_name)
        short_name = cls._build_short_name(first_name, last_name)
        if id: model["bag_key"] = cls._get_bag_key(id)
        if full_name: model["full_name"] = full_name
        if short_name: model["short_name"] = short_name

    @classmethod
    def _get_bag_key(cls, id):
        _bag = bag.Bag.get(account = id, raise_e = False)
        if not _bag: return None
        return _bag.key

    @classmethod
    def _build_full_name(cls, first_name, last_name):
        first_name = first_name or ""
        last_name = last_name or ""
        return first_name + (" " + last_name if last_name else "")

    @classmethod
    def _build_short_name(cls, first_name, last_name, limit = 16):
        first_name = first_name or ""
        last_name = last_name or ""
        first = first_name.split(" ")[0].strip()
        last = last_name.split(" ")[-1].strip()
        short_name = first + (" " + last if last else "")
        if len(short_name) <= limit: return short_name
        last = last_name[0] + "." if last_name else ""
        return first + (" " + last if last else "")

    @classmethod
    @appier.operation(
        name = "Import Social CSV",
        parameters = (
            ("CSV File", "file", "file"),
            ("Strict", "strict", bool, False)
        )
    )
    def import_social_csv_s(cls, file, strict):

        def callback(line):
            username, facebook_id, google_id = line
            account = cls.get(username = username, raise_e = strict)
            if not account: return
            if facebook_id: account.facebook_id = facebook_id
            if google_id: account.google_id = google_id
            account.save()

        cls._csv_import(file, callback)

    @classmethod
    @appier.link(name = "Export Simple")
    def simple_csv_url(cls, absolute = False):
        return appier.get_app().url_for(
            "account_api.simple_csv",
            absolute = absolute
        )

    def pre_create(self):
        appier_extras.admin.Account.pre_create(self)
        if not hasattr(self, "first_name") or not self.first_name:
            self.first_name = self.username

    def post_create(self):
        appier_extras.admin.Account.post_create(self)
        self.ensure_bag_s()
        self.notify()

    def recover_s(self, send_email = False):
        result = appier_extras.admin.Account.recover_s(
            self,
            send_email = send_email
        )
        self.notify(name = "account.recover")
        return result

    def ensure_bag_s(self, key = None):
        _bag = self.get_bag()
        if _bag: return _bag
        _bag = bag.Bag(key = key)
        _bag.account = self
        _bag.save()
        return _bag

    def get_bag(self):
        return bag.Bag.get(account = self.id, raise_e = False)

    def get_store_orders(self, status = "waiting_payment"):
        if not self.store: raise appier.OperationalError(
            message = "No store associated with user",
            code = 403
        )
        kwargs = dict()
        if self.store.is_restricted: kwargs["account"] = self.id
        orders = order.Order.find(
            status = status,
            store = self.store.id,
            sort = [("id", -1)],
            **kwargs
        )
        return orders

    @appier.operation(
        name = "Set Store",
        parameters = (
            (
                "Store",
                "store",
                appier.reference("Store", name = "id")
            ),
        )
    )
    def set_store_s(self, store):
        if not store: return
        self.store = store
        self.save()

    @appier.operation(
        name = "Set Metadata",
        parameters = (("Metadata","meta", "longmap"),)
    )
    def set_meta_s(self, meta):
        if not meta: return
        self.meta = meta
        self.save()

    @appier.operation(name = "Notify")
    def notify(self, name = None, *args, **kwargs):
        name = name or "account.new"
        account = self.reload(rules = False, map = True)
        receiver = account.get("email", None)
        receiver = kwargs.get("email", receiver)
        appier_extras.admin.Event.notify_g(
            name,
            arguments = dict(
                params = dict(
                    payload = account,
                    account = account,
                    receiver = receiver,
                    extra = kwargs
                )
            )
        )

    @appier.view(name = "Orders")
    def orders_v(self, *args, **kwargs):
        kwargs["sort"] = kwargs.get("sort", [("created", -1)])
        kwargs.update(account = self.id)
        return appier.lazy_dict(
            model = order.Order,
            kwargs = kwargs,
            entities = appier.lazy(lambda: order.Order.find(*args, **kwargs)),
            page = appier.lazy(lambda: order.Order.paginate(*args, **kwargs)),
            names = ["reference", "created", "total", "currency", "status"]
        )

    @appier.view(name = "Addresses")
    def addresses_v(self, *args, **kwargs):
        return appier.lazy_dict(
            model = self.addresses._target,
            kwargs = kwargs,
            entities = appier.lazy(lambda: self.addresses.find(*args, **kwargs)),
            page = appier.lazy(lambda: self.addresses.paginate(*args, **kwargs)),
            names = ["first_name", "last_name", "address", "country"]
        )

    @property
    def title(self):
        """
        Retrieves the title of the account, this value should
        be used as a prefix for the full name generation.

        :rtype: String
        :return: The title value for the current account, taking
        into account the selected gender.
        """

        return "Mrs." if self.gender == "Female" else "Mr."



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import logging
import unittest

import appier

import budy

class ProductTest(unittest.TestCase):

    def setUp(self):
        self.app = budy.BudyApp(level = logging.ERROR)

    def tearDown(self):
        self.app.unload()
        adapter = appier.get_adapter()
        adapter.drop_db()

    def test_add_remove_images(self):
        file = appier.typesf.ImageFile(
            dict(
                name = "name",
                data = "data",
                mime = "mime"
            )
        )

        media_1 = budy.Media(
            description = "description",
            label = "label",
            order = 1,
            file = file
        )
        media_1.save()

        media_2 = budy.Media(
            description = "description",
            label = "label",
            order = 1,
            file = file
        )
        media_2.save()

        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0
        )
        product.save()

        self.assertEqual(len(product.images), 0)

        product.add_image_s(media_1)

        self.assertEqual(len(product.images), 1)
        self.assertEqual(product.images[0].id, media_1.id)

        product.add_image_s(media_1)

        self.assertEqual(len(product.images), 1)
        self.assertEqual(product.images[0].id, media_1.id)

        product.add_image_s(media_2)

        self.assertEqual(len(product.images), 2)
        self.assertEqual(product.images[0].id, media_1.id)
        self.assertEqual(product.images[1].id, media_2.id)

        product = product.reload()

        self.assertEqual(len(product.images), 2)
        self.assertEqual(product.images[0].id, media_1.id)
        self.assertEqual(product.images[1].id, media_2.id)

        product.remove_image_s(media_1)

        self.assertEqual(len(product.images), 1)
        self.assertEqual(product.images[0].id, media_2.id)

        product.remove_image_s(media_1)

        self.assertEqual(len(product.images), 1)
        self.assertEqual(product.images[0].id, media_2.id)

        product.remove_image_s(media_2)

        self.assertEqual(len(product.images), 0)

        product = product.reload()

        self.assertEqual(len(product.images), 0)

    def test_measurements(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            quantity_hand = None
        )
        product.save()

        measurement = budy.Measurement(
            name = "size",
            value = 12,
            value_s = "12",
            price = None,
            quantity_hand = 2.0,
            product = product
        )
        measurement.save()

        self.assertEqual(product.quantity_hand, None)
        self.assertEqual(measurement.product.id, 1)
        self.assertEqual(measurement.product.id, product.id)

        product.measurements.append(measurement)
        product.save()

        self.assertEqual(product.quantity_hand, 2.0)
        self.assertEqual(len(product.measurements), 1)

        measurement = measurement.reload()

        self.assertEqual(measurement.product.id, 1)
        self.assertEqual(measurement.product.id, product.id)
        self.assertEqual(len(measurement.product.measurements), 1)

        result = product.get_measurement(12, name = "size")

        self.assertEqual(result.id, 1)
        self.assertEqual(result.name, "size")
        self.assertEqual(result.value, 12)
        self.assertEqual(result.value_s, "12")
        self.assertEqual(result.price, None)
        self.assertEqual(result.quantity_hand, 2.0)
        self.assertEqual(result.product.id, 1)
        self.assertEqual(result.product.short_description, "product")
        self.assertEqual(result.product.quantity_hand, 2.0)
        self.assertEqual(len(result.product.measurements), 1)

        measurement.delete()

        product = product.reload()

        self.assertEqual(len(product.measurements), 0)

    def test_labels(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            quantity_hand = None
        )
        product.save()

        collection = budy.Collection(
            name = "collection",
            new_in = True
        )
        collection.save()

        self.assertEqual(product.labels, [])
        self.assertEqual(len(product.collections), 0)
        self.assertEqual(collection.new_in, True)
        self.assertEqual(collection.labels, ["new_in"])

        product.add_collection_s(collection)

        self.assertEqual(product.labels, ["new_in"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)

        product.save()

        self.assertEqual(product.labels, ["new_in"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)

        collection.labels.append("extra")
        collection.save()

        self.assertEqual(product.labels, ["new_in"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)
        self.assertEqual(collection.new_in, True)
        self.assertEqual(collection.labels, ["new_in", "extra"])

        product = product.reload()
        product.save()

        self.assertEqual(product.labels, ["new_in", "extra"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)

        collection.labels.remove("extra")
        collection.save()

        product = product.reload()
        product.save()

        self.assertEqual(product.labels, ["new_in"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)

        collection.exclusive = True
        collection.save()

        self.assertEqual(product.labels, ["new_in"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)
        self.assertEqual(collection.new_in, True)
        self.assertEqual(collection.exclusive, True)
        self.assertEqual(collection.labels, ["new_in", "exclusive"])

        product = product.reload()
        product.save()

        self.assertEqual(product.labels, ["new_in", "exclusive"])
        self.assertEqual(len(product.collections), 1)
        self.assertEqual(product.collections[0].id, collection.id)

    def test_discount(self):
        product = budy.Product(
            short_description = "product",
            gender = "Male",
            price = 10.0,
            quantity_hand = None
        )
        product.save()

        self.assertEqual(product.price, 10.0)
        self.assertEqual(product.price_compare, 0.0)
        self.assertEqual(product.discount, 0.0)
        self.assertEqual(product.discount_percent, 0.0)
        self.assertEqual(product.is_discounted, False)

        product.price_compare = 16.0
        product.save()

        product = product.reload()

        self.assertEqual(product.price, 10.0)
        self.assertEqual(product.price_compare, 16.0)
        self.assertEqual(product.discount, 6.0)
        self.assertEqual(product.discount_percent, 37.5)
        self.assertEqual(product.is_discounted, True)

        product.price_compare = 20.0
        product.save()

        product = product.reload()

        self.assertEqual(product.price, 10.0)
        self.assertEqual(product.price_compare, 20.0)
        self.assertEqual(product.discount, 10.0)
        self.assertEqual(product.discount_percent, 50.0)
        self.assertEqual(product.is_discounted, True)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Budy
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Budy.
#
# Hive Budy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Budy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Budy. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from . import base

class Country(base.BudyBase):

    name = appier.field(
        index = True,
        default = True
    )

    country_code = appier.field(
        index = True
    )

    currency_code = appier.field(
        index = True
    )

    locale = appier.field(
        index = True
    )

    @classmethod
    def validate(cls):
        return super(Country, cls).validate() + [
            appier.not_null("name"),
            appier.not_empty("name"),

            appier.not_null("country_code"),
            appier.not_empty("country_code"),
            appier.string_eq("country_code", 2),

            appier.not_null("currency_code"),
            appier.not_empty("currency_code"),
            appier.string_eq("currency_code", 3),

            appier.not_null("locale"),
            appier.not_empty("locale"),
            appier.string_eq("locale", 5),
        ]

    @classmethod
    def list_names(cls):
        return ["id", "name", "country_code", "currency_code", "locale"]

    @classmethod
    def get_by_code(cls, country_code, *args, **kwargs):
        return cls.get(country_code = country_code, *args, **kwargs)

    @classmethod
    @appier.operation(
        name = "Import CSV",
        parameters = (
            ("CSV File", "file", "file"),
            ("Empty source", "empty", bool, False)
        )
    )
    def import_csv_s(cls, file, empty):

        def callback(line):
            name, country_code, currency_code, locale = line
            name = name or None
            country_code = country_code or None
            currency_code = currency_code or None
            locale = locale or None
            country = cls(
                name = name,
                country_code = country_code,
                currency_code = currency_code,
                locale = locale
            )
            country.save()

        if empty: cls.delete_c()
        cls._csv_import(file, callback)

    @classmethod
    @appier.link(name = "Export Simple")
    def simple_csv_url(cls, absolute = False):
        return appier.get_app().url_for(
            "country_api.simple_csv",
            absolute = absolute
        )

    @classmethod
    def _plural(cls):
        return "Countries"
